const createTestServer = require('./helpers/createTestServer')
const { Team, TeamMember, User } = require('..')

const clearDb = require('./_clearDb')

describe('Team API', () => {
  beforeEach(() => clearDb())

  afterAll(() => {
    const knex = Team.knex()
    knex.destroy()
  })

  it('returns global teams with members', async () => {
    const globalTeam = await Team.insert({
      role: 'editor',
      displayName: 'Editor',
      global: true,
    })

    const user = await User.insert({})

    const member = await TeamMember.insert({
      teamId: globalTeam.id,
      userId: user.id,
    })

    const GET_GLOBAL_TEAMS = `
      query GetGlobalTeams {
        getGlobalTeams {
          result {
            id
            members {
              id
              user {
                id
              }
            }
          }
          totalCount
        }
      }
    `

    const testServer = await createTestServer()

    const result = await testServer.executeOperation({
      query: GET_GLOBAL_TEAMS,
    })

    const data = result.data.getGlobalTeams

    expect(data.totalCount).toBe(1)
    expect(data.result).toHaveLength(1)
    const foundTeam = data.result[0]
    expect(foundTeam.id).toEqual(globalTeam.id)

    expect(foundTeam.members).toHaveLength(1)
    const foundMember = foundTeam.members[0]
    expect(foundMember.id).toEqual(member.id)

    const foundUser = foundMember.user
    expect(foundUser.id).toEqual(user.id)
  })

  it('returns only current user in members array when currentUserOnly flag is on', async () => {
    const globalTeam = await Team.insert({
      role: 'editor',
      displayName: 'Editor',
      global: true,
    })

    const anotherTeam = await Team.insert({
      role: 'author',
      displayName: 'Author',
      global: true,
    })

    const user = await User.insert({})
    const user2 = await User.insert({})

    const member = await TeamMember.insert({
      teamId: globalTeam.id,
      userId: user.id,
    })

    await TeamMember.insert({
      teamId: globalTeam.id,
      userId: user2.id,
    })

    // add user to two teams to make sure you get the correct member in the result
    await TeamMember.insert({
      teamId: anotherTeam.id,
      userId: user.id,
    })

    const GET_GLOBAL_TEAMS = `
      query GetGlobalTeams {
        getGlobalTeams {
          result {
            id
            role
            members(currentUserOnly:true) {
              id
              user {
                id
              }
            }
          }
          totalCount
        }
      }
    `

    const testServer = await createTestServer(user.id)

    const result = await testServer.executeOperation({
      query: GET_GLOBAL_TEAMS,
    })

    const foundTeam = result.data.getGlobalTeams.result.find(
      t => t.role === 'editor',
    )

    expect(foundTeam.id).toEqual(globalTeam.id)

    expect(foundTeam.members).toHaveLength(1)
    const foundMember = foundTeam.members[0]
    expect(foundMember.id).toEqual(member.id)

    const foundUser = foundMember.user
    expect(foundUser.id).toEqual(user.id)
  })
})
